from urllib.parse import quote, unquote


def gen_headers():
    return {
        "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/90.0.4430.85 Safari/537.36",
        "Referer": "https://www.bilibili.com",
    }


def gen_cookies(sessdata: str):
    # 先解码后编码是防止获取到的 SESSDATA 是已经解码后的（包含「,」）
    # 而番剧无法使用解码后的 SESSDATA
    return {"SESSDATA": quote(unquote(sessdata))}



import asyncio
import aiohttp
import random
from typing import Any, Optional

from aiohttp import ClientSession

from yutto.utils.file_buffer import AsyncFileBuffer
from yutto.utils.console.logger import Logger


class MaxRetryError(Exception):
    pass


class Fetcher:
    @classmethod
    async def fetch_text(cls, session: ClientSession, url: str, max_retry: int = 2) -> str:
        retry = max_retry + 1
        while retry:
            try:
                async with session.get(url) as resp:
                    return await resp.text()
            except asyncio.TimeoutError as e:
                Logger.warning("url: {url} 抓取超时".format(url=url))
            finally:
                retry -= 1
        raise MaxRetryError()

    @classmethod
    async def fetch_json(cls, session: ClientSession, url: str, max_retry: int = 2) -> Any:
        retry = max_retry + 1
        while retry:
            try:
                async with session.get(url) as resp:
                    return await resp.json()
            except asyncio.TimeoutError as e:
                Logger.warning("url: {url} 抓取超时".format(url=url))
            finally:
                retry -= 1
        raise MaxRetryError()

    @classmethod
    async def get_size(cls, session: ClientSession, url: str) -> Optional[int]:
        headers = session.headers.copy()
        headers["Range"] = "bytes=0-1"
        async with session.get(url, headers=headers) as resp:
            if resp.status == 206:
                return int(resp.headers["Content-Range"].split("/")[-1])
            else:
                return None

    @classmethod
    async def download_file_with_offset(
        cls,
        session: ClientSession,
        url: str,
        mirrors: list[str],
        file_buffer: AsyncFileBuffer,
        offset: int,
        size: Optional[int],
        stream: bool = True,
    ) -> None:
        done = False
        headers = session.headers.copy()
        url_pool = [url] + mirrors
        block_offset = 0
        while not done:
            try:
                url = random.choice(url_pool)
                headers["Range"] = "bytes={}-{}".format(
                    offset + block_offset, offset + size - 1 if size is not None else ""
                )
                async with session.get(
                    url, headers=headers, timeout=aiohttp.ClientTimeout(connect=5, sock_read=10)
                ) as resp:
                    if stream:
                        while True:
                            # 如果直接用 1KiB 的话，会产生大量的块，消耗大量的 CPU 资源，
                            # 反而使得协程的优势不明显
                            # 而使用 1MiB 以上或者不使用流式下载方式时，由于分块太大，
                            # 导致进度条显示的实时速度并不准，波动太大，用户体验不佳，
                            # 因此取两者折中
                            chunk = await resp.content.read(2 ** 15)
                            if not chunk:
                                break
                            await file_buffer.write(chunk, offset + block_offset)
                            block_offset += len(chunk)
                    else:
                        chunk = await resp.read()
                        await file_buffer.write(chunk, offset + block_offset)
                        block_offset += len(chunk)
                # TODO: 是否需要校验总大小
                done = True

            except asyncio.TimeoutError as e:
                Logger.warning("文件 {} 下载超时，尝试重新连接...".format(file_buffer.file_path))



import argparse
import asyncio
import os
import sys

import aiohttp

from yutto.api.info import is_vip
from yutto.media.codec import audio_codec_priority_default, video_codec_priority_default
from yutto.processor.crawler import gen_cookies, gen_headers
from yutto.utils.asynclib import install_uvloop
from yutto.utils.console.colorful import set_no_color
from yutto.utils.console.logger import Badge, Logger, set_logger_debug
from yutto.utils.ffmpeg import FFmpeg
from yutto.processor.filter import check_episodes


def check_basic_options(args: argparse.Namespace):
    """ 检查 argparse 无法检查的选项，并设置某些全局变量 """

    ffmpeg = FFmpeg()

    # 在使用 --no-color 或者环境变量 NO_COLOR 非空时都应该不显示颜色
    # See also: https://no-color.org/
    if args.no_color or os.environ.get("NO_COLOR"):
        set_no_color()

    # debug 设置
    if args.debug:
        set_logger_debug()
    else:
        # 为保证协程错误栈的可读性，debug 模式不启用 uvloop
        install_uvloop()

    # vcodec 检查
    vcodec_splited = args.vcodec.split(":")
    if len(vcodec_splited) != 2:
        Logger.error("vcodec 参数值（{}）不满足要求（并非使用 : 分隔的值）".format(args.vcodec))
        sys.exit(1)
    video_download_codec, video_save_codec = vcodec_splited
    if video_download_codec not in video_codec_priority_default:
        Logger.error(
            "download_vcodec 参数值（{}）不满足要求（允许值：{{{}}}）".format(
                video_download_codec, ", ".join(video_codec_priority_default)
            )
        )
        sys.exit(1)
    if video_save_codec not in ffmpeg.video_encodecs + ["copy"]:
        Logger.error(
            "save_vcodec 参数值（{}）不满足要求（允许值：{{{}}}）".format(video_save_codec, ", ".join(ffmpeg.video_encodecs + ["copy"]))
        )
        sys.exit(1)

    # acodec 检查
    acodec_splited = args.acodec.split(":")
    if len(acodec_splited) != 2:
        Logger.error("acodec 参数值（{}）不满足要求（并非使用 : 分隔的值）".format(args.acodec))
        sys.exit(1)
    audio_download_codec, audio_save_codec = acodec_splited
    if audio_download_codec not in audio_codec_priority_default:
        Logger.error(
            "download_acodec 参数值（{}）不满足要求（允许值：{{{}}}）".format(
                audio_download_codec, ", ".join(audio_codec_priority_default)
            )
        )
        sys.exit(1)
    if audio_save_codec not in ffmpeg.audio_encodecs + ["copy"]:
        Logger.error(
            "save_acodec 参数值（{}）不满足要求（允许值：{{{}}}）".format(audio_save_codec, ", ".join(ffmpeg.audio_encodecs + ["copy"]))
        )
        sys.exit(1)

    # only_video 和 only_audio 不能同时设置
    if not args.require_video and not args.require_audio:
        Logger.error("only_video 和 only_audio 不能同时设置")
        sys.exit(1)

    # TODO: proxy 检验

    # 不下载视频无法嵌入字幕
    if not args.require_video and args.embed_subtitle:
        Logger.error("不下载视频时无法嵌入字幕")
        sys.exit(1)

    # 不下载视频无法嵌入弹幕
    if not args.require_video and args.embed_danmaku:
        Logger.error("不下载视频时无法嵌入弹幕")
        sys.exit(1)

    # 嵌入弹幕功能仅支持 ASS 弹幕
    if args.embed_danmaku and args.danmaku != "ass":
        Logger.error("嵌入弹幕功能仅支持 ASS 弹幕")
        sys.exit(1)

    # 大会员身份校验
    if not args.sessdata:
        Logger.warning("未提供 SESSDATA，无法下载会员专享剧集")
    elif asyncio.run(check_is_vip(args.sessdata)):
        Logger.custom("成功以大会员身份登录～", badge=Badge("大会员", fore="white", back="magenta"))
    else:
        Logger.warning("以非大会员身份登录，无法下载会员专享剧集")


def check_batch_options(args: argparse.Namespace):
    """ 检查批量下载相关选项 """
    # 检查 episodes 格式（简单的正则检查，后续过滤剧集时还有完整检查）
    if not check_episodes(args.episodes):
        Logger.error("选集参数（{}）格式不正确".format(args.episodes))
        sys.exit(1)


async def check_is_vip(sessdata: str = "") -> bool:
    async with aiohttp.ClientSession(
        headers=gen_headers(), cookies=gen_cookies(sessdata), timeout=aiohttp.ClientTimeout(total=5)
    ) as session:
        return await is_vip(session)



import re
from typing import Literal
from urllib.parse import unquote

_count: int = 0


def size_format(size: float, ndigits: int = 2, baseUnitSize: Literal[1024, 1000] = 1024) -> str:
    """ 输入数据字节数，与保留小数位数，返回数据量字符串 """
    sign = "-" if size < 0 else ""
    size = abs(size)
    unit_list = (
        ["Bytes", "KiB", "MiB", "GiB", "TiB", "PiB", "EiB", "ZiB", "YiB", "BiB"]
        if baseUnitSize == 1024
        else ["Bytes", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB", "BB"]
    )

    index = 0
    while index < len(unit_list) - 1:
        if size >= baseUnitSize ** (index + 1):
            index += 1
        else:
            break
    return "{}{:.{}f} {}".format(sign, size / baseUnitSize ** index, ndigits, unit_list[index])


def get_char_width(char: str) -> int:
    """ 计算单个字符的宽度 """
    # fmt: off
    widths = [
        (126, 1), (159, 0), (687, 1), (710, 0), (711, 1),
        (727, 0), (733, 1), (879, 0), (1154, 1), (1161, 0),
        (4347, 1), (4447, 2), (7467, 1), (7521, 0), (8369, 1),
        (8426, 0), (9000, 1), (9002, 2), (11021, 1), (12350, 2),
        (12351, 1), (12438, 2), (12442, 0), (19893, 2), (19967, 1),
        (55203, 2), (63743, 1), (64106, 2), (65039, 1), (65059, 0),
        (65131, 2), (65279, 1), (65376, 2), (65500, 1), (65510, 2),
        (120831, 1), (262141, 2), (1114109, 1),
    ]
    # fmt: on

    o = ord(char)
    if o == 0xE or o == 0xF:
        return 0
    for num, wid in widths:
        if o <= num:
            return wid
    return 1


def get_string_width(string: str) -> int:
    """ 计算包含中文的字符串宽度 """
    # 去除颜色码
    string = no_color_string(string)
    try:
        length = sum([get_char_width(c) for c in string])
    except:
        length = len(string)
    return length


def no_color_string(string: str) -> str:
    """ 去除字符串中的颜色码 """
    regex_color = re.compile(r"\033\[\d+m")
    string = regex_color.sub("", string)
    return string


def repair_filename(filename: str) -> str:
    """ 修复不合法的文件名 """

    def to_full_width_chr(matchobj: "re.Match[str]") -> str:
        char = matchobj.group(0)
        full_width_char = chr(ord(char) + ord("？") - ord("?"))
        return full_width_char

    # 路径非法字符，转全角
    regex_path = re.compile(r'[\\/:*?"<>|]')
    # 空格类字符，转空格
    regex_spaces = re.compile(r"\s+")
    # 不可打印字符，移除
    regex_non_printable = re.compile(
        r"[\001\002\003\004\005\006\007\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f"
        r"\x10\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1a]"
    )

    # url decode
    filename = unquote(filename)
    filename = regex_path.sub(to_full_width_chr, filename)
    filename = regex_spaces.sub(" ", filename)
    filename = regex_non_printable.sub("", filename)
    filename = filename.strip()
    if not filename:
        filename = "未命名文件_{:04}".format(_count)
        _count += 1
    return filename



import asyncio
import os
import time
from typing import Any, Optional

import aiohttp
from aiofiles import os as aioos

from yutto.api.types import AudioUrlMeta, VideoUrlMeta
from yutto.processor.filter import filter_none_value, select_audio, select_video
from yutto.utils.asynclib import CoroutineTask, parallel_with_limit
from yutto.utils.console.logger import Logger
from yutto.utils.fetcher import Fetcher
from yutto.utils.ffmpeg import FFmpeg
from yutto.utils.file_buffer import AsyncFileBuffer
from yutto.processor.progressor import show_progress


def slice(start: int, total_size: Optional[int], block_size: Optional[int] = None) -> list[tuple[int, Optional[int]]]:
    """生成分块后的 (start, size) 序列

    Args:
        start (int): 总起始位置
        total_size (Optional[int]): 需要分块的总大小
        block_size (Optional[int], optional): 每块的大小. Defaults to None.

    Returns:
        list[tuple[int, Optional[int]]]: 分块大小序列，使用元组组织，格式为 (start, size)
    """
    if total_size is None:
        return [(0, None)]
    if block_size is None:
        return [(0, total_size - 1)]
    assert start <= total_size, "起始地址（{}）大于总地址（{}）".format(start, total_size)
    offset_list: list[tuple[int, Optional[int]]] = [(i, block_size) for i in range(start, total_size, block_size)]
    if (total_size - start) % block_size != 0:
        offset_list[-1] = (
            start + (total_size - start) // block_size * block_size,
            total_size - start - (total_size - start) // block_size * block_size,
        )
    return offset_list


def combine(*l_list: list[Any]) -> list[Any]:
    """将多个 list 「均匀」地合并到一个 list

    # example

    ```
    l_list = [
        [1, 2, 3, 4, 5],
        [6, 7, 8],
        [9, 10, 11, 12]
    ]
    combine(l_list)
    # [1, 6, 9, 2, 7, 10, 3, 8, 11, 4, 12, 5]
    ```
    """
    results: list[Any] = []
    for i in range(max([len(l) for l in l_list])):
        for l in l_list:
            if i < len(l):
                results.append(l[i])
    return results


async def download_video(
    session: aiohttp.ClientSession,
    videos: list[VideoUrlMeta],
    audios: list[AudioUrlMeta],
    output_dir: str,
    file_name: str,
    # TODO: options 使用 TypedDict
    options: Any,
):
    video_path = os.path.join(output_dir, file_name + "_video.m4s")
    audio_path = os.path.join(output_dir, file_name + "_audio.m4s")
    output_path_template = os.path.join(output_dir, file_name + "{output_format}")
    ffmpeg = FFmpeg()

    # TODO: 显示全部 Videos、Audios 信息
    video = select_video(videos, options["require_video"], options["video_quality"], options["video_download_codec"])
    audio = select_audio(audios, options["require_audio"], options["audio_quality"], options["audio_download_codec"])
    # TODO: 显示被选中的 Video、Audio 信息

    output_format = ".mp4" if video is not None else ".aac"
    output_path = output_path_template.format(output_format=output_format)
    if not options["overwrite"] and os.path.exists(output_path):
        Logger.info("文件 {} 已存在".format(file_name))
        return

    # idx_video = -1
    # if video is not None:
    #     idx_video = videos.index(video)
    # Logger.info(f"视频 {file_name} 共包含以下 {len(videos)} 个视频流：")
    # videos_log = [
    #     "{:02} [{:>4}] [{:>4}x{:>4}] <{:>10}>".format(
    #         i,
    #         video["codec"].upper(),
    #         video["width"],
    #         video["height"],
    #         video_quality_map[video["quality"]]["description"],
    #     )
    #     for i, video in enumerate(videos)
    # ]

    # for video_log in videos_log:
    #     Logger.info(video_log)

    if video is None and audio is None:
        return
    buffers: list[Optional[AsyncFileBuffer]] = [None, None]
    sizes: list[Optional[int]] = [None, None]
    task_funcs: list[list[CoroutineTask]] = []
    if video is not None:
        vbuf = await AsyncFileBuffer.create(video_path, overwrite=options["overwrite"])
        vsize = await Fetcher.get_size(session, video["url"])
        vtask_funcs = [
            Fetcher.download_file_with_offset(session, video["url"], video["mirrors"], vbuf, offset, block_size)
            for offset, block_size in slice(vbuf.written_size, vsize, options["block_size"])
        ]
        task_funcs.append(vtask_funcs)
        buffers[0], sizes[0] = vbuf, vsize

    if audio is not None:
        abuf = await AsyncFileBuffer.create(audio_path, overwrite=options["overwrite"])
        asize = await Fetcher.get_size(session, audio["url"])
        atask_funcs = [
            Fetcher.download_file_with_offset(session, audio["url"], audio["mirrors"], abuf, offset, block_size)
            for offset, block_size in slice(abuf.written_size, asize, options["block_size"])
        ]
        task_funcs.append(atask_funcs)
        buffers[1], sizes[1] = abuf, asize

    tasks = parallel_with_limit(combine(*task_funcs), num_workers=options["num_workers"])
    tasks.append(asyncio.create_task(show_progress(filter_none_value(buffers), sum(filter_none_value(sizes)))))

    Logger.info(f"开始下载 {file_name}……")
    for task in tasks:
        await task
    Logger.info("下载完成！")

    if video is not None:
        await vbuf.close()
    if audio is not None:
        await abuf.close()

    # TODO: 将 merge 分离出去？
    Logger.info(f"开始合并 {file_name}……")
    # fmt: off
    args: list[str] = []
    if video is not None:
        args.extend([
            "-i", video_path,
        ])
    if audio is not None:
        args.extend([
            "-i", audio_path,
        ])
    if video is not None:
        args.extend([
            "-vcodec", options["video_save_codec"],
        ])
    if audio is not None:
        args.extend([
            "-acodec", options["audio_save_codec"],
        ])
    args.extend(["-y"])

    args.append(output_path)
    ffmpeg.exec(args)
    # fmt: on
    Logger.info("合并完成！")

    if video is not None:
        await aioos.remove(video_path)
    if audio is not None:
        await aioos.remove(audio_path)



import argparse

from yutto.cli import get, info, batch_get, check_options
from yutto.__version__ import __version__
from yutto.utils.ffmpeg import FFmpeg
from yutto.utils.console.colorful import colored_string
from yutto.utils.console.logger import Logger
from yutto.media.quality import video_quality_priority_default, audio_quality_priority_default


def main():
    parser = argparse.ArgumentParser(description="yutto 一个可爱且任性的 B 站视频下载器", prog="yutto")
    parser.add_argument("-v", "--version", action="version", version="%(prog)s {}".format(__version__))
    parser.add_argument("-n", "--num-workers", type=int, default=8, help="同时用于下载的最大 Worker 数")
    parser.add_argument(
        "-q",
        "--video-quality",
        default=125,
        choices=video_quality_priority_default,
        type=int,
        help="视频清晰度等级（125:HDR, 120:4K, 116:1080P60, 112:1080P+, 80:1080P, 74:720P60, 64:720P, 32:480P, 16:360P）",
    )
    parser.add_argument(
        "--audio-quality",
        default=30280,
        choices=audio_quality_priority_default,
        type=int,
        help="音频码率等级（30280:320kbps, 30232:128kbps, 30216:64kbps）",
    )
    parser.add_argument("--vcodec", default="avc:copy", help="视频编码格式（<下载格式>:<生成格式>）")
    parser.add_argument("--acodec", default="mp4a:copy", help="音频编码格式（<下载格式>:<生成格式>）")
    parser.add_argument("--only-video", dest="require_audio", action="store_false", help="只下载视频")
    parser.add_argument("--only-audio", dest="require_video", action="store_false", help="只下载音频")
    parser.add_argument("--danmaku", default="xml", choices=["xml", "ass", "no"], help="弹幕类型，no 为不生成弹幕")
    parser.add_argument("-b", "--block-size", default=0.5, type=float, help="分块下载时各块大小，单位为 MiB，默认为 0.5MiB")
    parser.add_argument("-w", "--overwrite", action="store_true", help="强制覆盖已下载内容")
    parser.add_argument("-x", "--proxy", default="auto", help="设置代理（auto 为系统代理、no 为不使用代理、当然也可以设置代理值）")
    parser.add_argument("-d", "--dir", default="", help="下载目录")
    parser.add_argument("-c", "--sessdata", default="", help="Cookies 中的 SESSDATA 字段")
    parser.add_argument("--path-pattern", default="{auto}", help="多级目录的存储路径 Pattern")
    parser.add_argument("--no-subtitle", action="store_true", help="不生成字幕文件")
    parser.add_argument("--embed-danmaku", action="store_true", help="将弹幕文件嵌入到视频中")
    parser.add_argument("--embed-subtitle", default=None, help="将字幕文件嵌入到视频中（需输入语言代码）")
    parser.add_argument("--no-color", action="store_true", help="不使用颜色")
    parser.add_argument("--debug", action="store_true", help="启用 debug 模式")
    parser.set_defaults(action=run)

    subparsers = parser.add_subparsers()
    # 子命令 get
    parser_get = subparsers.add_parser("get", help="获取单个视频")
    get.add_get_arguments(parser_get)
    # 子命令 info
    # TODO
    # 子命令 batch
    parser_batch = subparsers.add_parser("batch", help="批量获取视频")
    subparsers_batch = parser_batch.add_subparsers()
    # 子命令 batch get
    parser_batch_get = subparsers_batch.add_parser("get", help="批量获取视频")
    batch_get.add_get_arguments(parser_batch_get)
    # 执行各自的 action
    args = parser.parse_args()
    check_options.check_basic_options(args)
    args.action(args)


def run(args: argparse.Namespace):
    Logger.error("未指定子命令 (get, info, batch)")
    Logger.info("yutto version: {}".format(colored_string(__version__, fore="green")))
    Logger.info("FFmpeg version: {}".format(colored_string(FFmpeg().version, fore="blue")))


if __name__ == "__main__":
    main()



import asyncio
import time

from yutto.utils.console.formatter import size_format
from yutto.utils.file_buffer import AsyncFileBuffer
from yutto.utils.console.logger import Logger


async def show_progress(file_buffers: list[AsyncFileBuffer], total_size: int):
    file_buffers = list(filter(lambda x: x is not None, file_buffers))
    t = time.time()
    size = sum([file_buffer.written_size for file_buffer in file_buffers])
    while True:
        size_in_buffer: int = sum(
            [sum([len(chunk.data) for chunk in file_buffer.buffer]) for file_buffer in file_buffers]
        )
        size_written: int = sum([file_buffer.written_size for file_buffer in file_buffers])

        t_now = time.time()
        size_now = size_written + size_in_buffer
        speed = (size_now - size) / (t_now - t + 10 ** -6)

        Logger.print(
            "[File: {:>10} + Buffer: {:>10}({:>4} 块)]/{:>10} {:>10}/s".format(
                size_format(size_written),
                size_format(size_in_buffer),
                sum([len(file_buffer.buffer) for file_buffer in file_buffers]),
                size_format(total_size),
                size_format(speed),
            ),
            end="\r",
        )
        t, size = t_now, size_now
        await asyncio.sleep(0.5)
        if total_size == size:
            break
